import time

from nyckel import ClassificationAnnotation, ClassificationLabel, TextClassificationFunction, TextClassificationSample


def test_simple(text_classification_function_with_content: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function_with_content

    samples = func.list_samples()
    assert len(samples) > 0
    func.delete_samples([sample.id for sample in samples if sample.id])
    time.sleep(2)
    samples = func.list_samples()
    assert len(samples) == 0


def test_duplicate_samples(text_classification_function: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function

    samples = [
        TextClassificationSample(data="hello"),
        TextClassificationSample(data="hello"),
    ]

    sample_ids = func.create_samples(samples)
    assert len(sample_ids) == 2
    assert sample_ids[0] == sample_ids[1]


def test_update_annotation(text_classification_function: TextClassificationFunction) -> None:
    func: TextClassificationFunction = text_classification_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Bad")]
    nice = ClassificationAnnotation(label_name="Nice")
    bad = ClassificationAnnotation(label_name="Bad")
    sample = TextClassificationSample(data="Hello, friend", external_id="1", annotation=nice)

    # Create assets and make sure the annotation is set correctly.
    func.create_labels(labels_to_create)
    sample_id = func.create_samples([sample])[0]
    time.sleep(1)
    sample = func.read_sample(sample_id)
    assert sample.annotation and sample.annotation.label_name == "Nice"

    # Remove the annotation, update it and make sure it's updated.
    sample.annotation = None
    func.update_annotation(sample)
    time.sleep(1)
    sample = func.read_sample(sample_id)
    assert not sample.annotation

    # Change the annotation, update it and make sure it's updated.
    sample.annotation = bad
    func.update_annotation(sample)
    time.sleep(1)
    sample = func.read_sample(sample_id)
    assert sample.annotation and sample.annotation.label_name == "Bad"


def test_list_order(text_classification_function: TextClassificationFunction) -> None:
    # Create a few samples
    samples = [TextClassificationSample(data=f"Sample {i}") for i in range(10)]
    _ = text_classification_function.create_samples(samples)
    time.sleep(1)
    # Create one new sample
    sample = TextClassificationSample(data="New Sample")
    new_sample_id = text_classification_function.create_samples([sample])[0]
    time.sleep(1)

    # List the samples and make sure the order is correct
    listed_samples = text_classification_function.list_samples()
    assert len(listed_samples) == 11
    assert listed_samples[0].id == new_sample_id



import base64
import os
from io import BytesIO
from typing import Tuple, Union

import requests
from PIL import Image

from nyckel.config import MAX_IMAGE_SIZE_PIXELS


class ImageResizer:

    def __init__(self, max_image_size_pixels: int = MAX_IMAGE_SIZE_PIXELS):
        self._max_image_size_pixels = max_image_size_pixels

    def __call__(self, img: Image.Image) -> Image.Image:
        if not self._needs_resize(img.width, img.height):
            return img
        new_width, new_height = self._get_new_width_height(img.width, img.height)
        img = img.resize((new_width, new_height))
        return img

    def _needs_resize(self, width: int, height: int) -> bool:
        return width > self._max_image_size_pixels or height > self._max_image_size_pixels

    def _get_new_width_height(self, width: int, height: int) -> Tuple[int, int]:
        if width > height:
            new_width = self._max_image_size_pixels
            new_height = int(new_width * height / width)
        else:
            new_height = self._max_image_size_pixels
            new_width = int(new_height * width / height)
        return new_width, new_height


class ImageDecoder:
    def to_image(self, sample_data: str) -> Image.Image:
        byte_stream = self.to_stream(sample_data)
        try:
            img = Image.open(byte_stream)
        except OSError:
            raise ValueError("Truncated Image Bytes")
        return img

    def to_stream(self, sample_data: str) -> BytesIO:
        if self.looks_like_url(sample_data):
            return self._load_from_url(sample_data)
        if self.looks_like_local_filepath(sample_data):
            return self._load_from_local_filepath(sample_data)
        if self.looks_like_data_uri(sample_data):
            return self._load_from_data_uri(sample_data)
        raise ValueError(f"Unable to parse input {sample_data=}.")

    def looks_like_url(self, sample_data: str) -> bool:
        return sample_data.startswith("https://") or sample_data.startswith("http://")

    def _load_from_url(self, url: str) -> BytesIO:
        response = requests.get(url, timeout=5)
        return BytesIO(response.content)

    def looks_like_local_filepath(self, local_path: str) -> bool:
        return os.path.exists(local_path)

    def _load_from_local_filepath(self, local_path: str) -> BytesIO:
        with open(local_path, "rb") as fh:
            return BytesIO(fh.read())

    def looks_like_data_uri(self, data_uri: str) -> bool:
        return data_uri.startswith("data:image")

    def _load_from_data_uri(self, data_uri: str) -> BytesIO:
        self._validate_image_data_uri(data_uri)
        image_b64_encoded_string = self.strip_base64_prefix(data_uri)
        im_bytes = base64.b64decode(image_b64_encoded_string)
        return BytesIO(im_bytes)

    def _validate_image_data_uri(self, inline_data: str) -> None:
        if inline_data == "":
            raise ValueError("Empty string")
        if "base64" not in inline_data:
            raise ValueError("base64 not in preamble.")
        inline_data_parts = inline_data.split(";base64,")
        if not len(inline_data_parts) == 2:
            raise ValueError("Unable to parse byte string.")
        if inline_data_parts[1] == "":
            raise ValueError("Empty image content")

    def strip_base64_prefix(self, inline_data: str) -> str:
        return inline_data.split(";base64,")[1]


class ImageEncoder:
    def to_base64(self, img: Union[Image.Image, BytesIO]) -> str:
        if isinstance(img, Image.Image):
            im_bytes = BytesIO()
            if img.mode == "P" and "transparency" in img.info:
                # Convert to RGBA if the image has transparency.
                img = img.convert("RGBA")
            if img.mode == "RGBA":
                # Explicitly set RGBA backgrounds to white.
                background = Image.new("RGBA", img.size, (255, 255, 255))
                background.paste(img, mask=img.split()[3])  # 3 is the alpha channel
                img = background
            if not img.mode == "RGB":
                img = img.convert("RGB")
            img.save(im_bytes, format="JPEG", quality=95)
        else:
            im_bytes = img
        encoded_string = base64.b64encode(im_bytes.getvalue()).decode("utf-8")
        return "data:image/jpg;base64," + encoded_string



from nyckel import ImageDecoder, ImageEncoder, ImageResizer, ImageSampleData
from nyckel.data_classes import NyckelId


def strip_nyckel_prefix(prefixed_id: str) -> str:
    split_id = prefixed_id.split("_")
    if len(split_id) == 2:
        return split_id[1]
    else:
        return prefixed_id


def is_nyckel_owned_url(url: str) -> bool:
    return url.startswith("https://s3.us-west-2.amazonaws.com/nyckel.server.")


class ImageSampleBodyTransformer:

    def __init__(self):
        self._decoder = ImageDecoder()
        self._encoder = ImageEncoder()
        self._resizer = ImageResizer()

    def __call__(self, sample_data: ImageSampleData) -> str:
        """Resizes if needed and encodes the sample data as a URL or dataURI."""
        if is_nyckel_owned_url(sample_data):
            # If the input points to a Nyckel S3 bucket, we know that the image is processed and verified.
            # In that case, we just point back to that URL.
            return sample_data

        if self._decoder.looks_like_url(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        if self._decoder.looks_like_local_filepath(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        if self._decoder.looks_like_data_uri(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        raise ValueError(f"Can't parse input sample.data={sample_data}")


class ImageFieldTransformer:

    def __init__(self, field_id: NyckelId):
        self._field_id = field_id
        self._decoder = ImageDecoder()
        self._encoder = ImageEncoder()
        self._resizer = ImageResizer(max_image_size_pixels=384)

    def __call__(self, tabular_sample_data: dict) -> dict:
        if self._field_id in tabular_sample_data:
            tabular_sample_data[self._field_id] = self._sample_data_to_body(tabular_sample_data[self._field_id])
        return tabular_sample_data

    def _sample_data_to_body(self, sample_data: str) -> str:
        """Resizes if needed and encodes the sample data as a URL or dataURI."""
        if is_nyckel_owned_url(sample_data):
            # If the input points to a Nyckel S3 bucket, we know that the image is processed and verified.
            # In that case, we just point back to that URL.
            return sample_data

        if self._decoder.looks_like_url(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        if self._decoder.looks_like_local_filepath(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        if self._decoder.looks_like_data_uri(sample_data):
            return self._encoder.to_base64(self._resizer(self._decoder.to_image(sample_data)))

        raise ValueError(f"Can't parse input sample.data={sample_data}")



import os
import time

from conftest import make_random_image
from nyckel import ClassificationLabel, ClassificationPrediction, ImageTagsFunction, ImageTagsSample, TagsAnnotation
from PIL import Image


class TestFunction:

    # Create & Delete happens in the test fixture, and there is no update operation
    # So remains to test read.

    def test_read(self, image_tags_function, auth_test_credentials):
        function_id = image_tags_function.function_id
        reloaded_function = ImageTagsFunction(function_id, auth_test_credentials)
        assert reloaded_function.function_id == function_id


class TestLabels:

    def test_create(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        label_ids = func.create_labels(["cat", "dog"])
        assert len(label_ids) == 2

    def test_read(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        label_id = func.create_labels(["dog"])[0]
        label = func.read_label(label_id)
        assert label.name == "dog"
        assert label.id == label_id

    def test_update(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        label_id = func.create_labels(["dog"])[0]
        updated_label = ClassificationLabel(name="doggo", id=label_id)
        func.update_label(updated_label)
        time.sleep(0.5)
        labels = func.list_labels()
        assert labels[0].name == "doggo"

    def test_delete(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        label_id = func.create_labels(["dog"])[0]
        func.delete_labels([label_id])
        time.sleep(0.5)
        labels = func.list_labels()
        assert len(labels) == 0


class TestSamples:

    def test_create_inline(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        sample_ids = func.create_samples([make_random_image(), make_random_image()])
        assert len(sample_ids) == 2

    def test_create_PIL(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        sample_ids = func.create_samples([Image.new("RGB", (100, 100)), Image.new("RGB", (80, 80))])
        assert len(sample_ids) == 2

    def test_create_local_path(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        local_filepath = os.path.abspath("tests/fixtures/flower.jpg")
        sample_ids = func.create_samples([local_filepath])
        assert len(sample_ids) == 1

    def test_create_url(self, image_tags_function: ImageTagsFunction):
        func = image_tags_function
        sample_ids = func.create_samples(["https://picsum.photos/200", "https://picsum.photos/300"])
        assert len(sample_ids) == 2


def test_end_to_end(image_tags_function):
    func: ImageTagsFunction = image_tags_function

    labels_to_create = [ClassificationLabel(name="Nice"), ClassificationLabel(name="Boo")]
    func.create_labels(labels_to_create)

    samples = [
        ImageTagsSample(data=make_random_image(), annotation=[TagsAnnotation("Nice")]),
        ImageTagsSample(data=make_random_image(), annotation=[TagsAnnotation("Nice"), TagsAnnotation("Boo")]),
        ImageTagsSample(data=make_random_image(), annotation=[TagsAnnotation("Boo")]),
        ImageTagsSample(data=make_random_image(), annotation=[TagsAnnotation("Boo")]),
        ImageTagsSample(data=make_random_image(), annotation=[]),
    ]

    func.create_samples(samples)

    while not func.has_trained_model():
        print("-> No trained model yet. Sleeping 1 sec...")
        time.sleep(1)

    predictions = func.invoke([make_random_image(), make_random_image(), make_random_image()])
    assert len(predictions) == 3
    for prediction in predictions:
        if len(prediction) > 0:
            assert isinstance(prediction[0], ClassificationPrediction)

    returned_samples = func.list_samples()
    assert len(samples) == len(returned_samples)
    assert isinstance(returned_samples[0], ImageTagsSample)



from typing import Dict, List, Sequence, Tuple, Union

from PIL import Image

from nyckel import (
    ClassificationAnnotation,
    ClassificationFunction,
    ClassificationLabel,
    ClassificationPrediction,
    Credentials,
    ImageClassificationSample,
    ImageEncoder,
    ImageSampleData,
    LabelName,
    NyckelId,
)
from nyckel.functions.classification import factory
from nyckel.functions.classification.classification import ClassificationFunctionURLHandler
from nyckel.functions.classification.function_handler import ClassificationFunctionHandler
from nyckel.functions.classification.label_handler import ClassificationLabelHandler
from nyckel.functions.classification.sample_handler import ClassificationSampleHandler
from nyckel.functions.utils import ImageSampleBodyTransformer, strip_nyckel_prefix


class ImageClassificationFunction(ClassificationFunction):
    """
    Example:

    ```py
    from nyckel import Credentials, ImageClassificationFunction

    credentials = Credentials(client_id="...", client_secret="...")

    func = ImageClassificationFunction.create("IsCatOrDog", credentials)
    func.create_samples([
        ("cat1.jpg", "cat"),
        ("cat2.jpg", "cat"),
        ("dog1.jpg", "dog"),
        ("dog2.jpg", "dog")
    ])

    predictions = func.invoke(["cat_or_dog.jpg"])
    ```
    """

    def __init__(self, function_id: str, credentials: Credentials) -> None:
        self._function_id = function_id
        self._function_handler = ClassificationFunctionHandler(function_id, credentials)
        self._label_handler = ClassificationLabelHandler(function_id, credentials)
        self._url_handler = ClassificationFunctionURLHandler(function_id, credentials.server_url)
        self._sample_handler = ClassificationSampleHandler(function_id, credentials)
        self._encoder = ImageEncoder()
        assert self._function_handler.get_input_modality() == "Image"

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        status_string = f"Name: {self.name}, id: {self.function_id}, url: {self._url_handler.train_page}"
        return status_string

    @property
    def function_id(self) -> NyckelId:
        return self._function_id

    @property
    def sample_count(self) -> int:
        return self._function_handler.sample_count

    @property
    def label_count(self) -> int:
        return self._function_handler.label_count

    @property
    def name(self) -> str:
        return self._function_handler.get_name()

    @classmethod
    def create(cls, name: str, credentials: Credentials) -> "ImageClassificationFunction":
        return factory.ClassificationFunctionFactory.create(name, "Image", credentials)  # type: ignore

    def delete(self) -> None:
        self._function_handler.delete()

    def invoke(  # type: ignore
        self, sample_data_list: List[ImageSampleData], model_id: str = ""
    ) -> List[ClassificationPrediction]:
        return self._sample_handler.invoke(sample_data_list, ImageSampleBodyTransformer(), model_id=model_id)

    def has_trained_model(self) -> bool:
        return self._function_handler.is_trained

    def create_labels(self, labels: Sequence[Union[ClassificationLabel, str]]) -> List[NyckelId]:
        typed_labels = [
            label if isinstance(label, ClassificationLabel) else ClassificationLabel(name=label) for label in labels
        ]
        return self._label_handler.create_labels(typed_labels)

    def list_labels(self) -> List[ClassificationLabel]:
        return self._label_handler.list_labels(self.label_count)

    def read_label(self, label_id: NyckelId) -> ClassificationLabel:
        return self._label_handler.read_label(label_id)

    def update_label(self, label: ClassificationLabel) -> ClassificationLabel:
        return self._label_handler.update_label(label)

    def delete_labels(self, label_ids: List[NyckelId]) -> None:
        return self._label_handler.delete_labels(label_ids)

    def create_samples(  # type: ignore
        self,
        samples: Sequence[  # type: ignore
            Union[
                ImageClassificationSample,
                Tuple[Image.Image, LabelName],
                Image.Image,
                Tuple[ImageSampleData, LabelName],
                ImageSampleData,
            ]
        ],
    ) -> List[NyckelId]:
        typed_samples = self._wrangle_post_samples_input(samples)
        typed_samples = self._strip_label_names(typed_samples)
        self._create_labels_as_needed(typed_samples)

        return self._sample_handler.create_samples(typed_samples, ImageSampleBodyTransformer())

    def list_samples(self) -> List[ImageClassificationSample]:  # type: ignore
        samples_dict_list = self._sample_handler.list_samples(self.sample_count)
        labels = self._label_handler.list_labels(None)

        label_name_by_id = {label.id: label.name for label in labels}

        return [self._sample_from_dict(entry, label_name_by_id) for entry in samples_dict_list]

    def read_sample(self, sample_id: NyckelId) -> ImageClassificationSample:
        sample_dict = self._sample_handler.read_sample(sample_id)

        labels = self._label_handler.list_labels(None)
        label_name_by_id = {label.id: label.name for label in labels}

        return self._sample_from_dict(sample_dict, label_name_by_id)

    def update_annotation(self, sample: ImageClassificationSample) -> None:  # type: ignore
        self._sample_handler.update_annotation(sample)

    def delete_samples(self, sample_ids: List[NyckelId]) -> None:
        self._sample_handler.delete_samples(sample_ids)

    def _sample_from_dict(self, sample_dict: Dict, label_name_by_id: Dict) -> ImageClassificationSample:
        if "annotation" in sample_dict:
            annotation = ClassificationAnnotation(
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["annotation"]["labelId"])],
            )
        else:
            annotation = None
        if "prediction" in sample_dict:
            prediction = ClassificationPrediction(
                confidence=sample_dict["prediction"]["confidence"],
                label_name=label_name_by_id[strip_nyckel_prefix(sample_dict["prediction"]["labelId"])],
            )
        else:
            prediction = None
        return ImageClassificationSample(
            id=strip_nyckel_prefix(sample_dict["id"]),
            data=sample_dict["data"],
            external_id=sample_dict["externalId"] if "externalId" in sample_dict else None,
            annotation=annotation,
            prediction=prediction,
        )

    def _wrangle_post_samples_input(
        self,
        samples: Sequence[
            Union[
                ImageSampleData,
                Image.Image,
                ImageClassificationSample,
                Tuple[ImageSampleData, LabelName],
                Tuple[Image.Image, LabelName],
            ]
        ],
    ) -> List[ImageClassificationSample]:
        typed_samples: List[ImageClassificationSample] = []
        for sample in samples:
            if isinstance(sample, str):
                typed_samples.append(ImageClassificationSample(data=sample))
            elif isinstance(sample, Image.Image):
                typed_samples.append(ImageClassificationSample(data=self._encoder.to_base64(sample)))
            elif isinstance(sample, (tuple, list)) and isinstance(sample[0], str):
                image_str, label_name = sample
                typed_samples.append(
                    ImageClassificationSample(
                        data=image_str, annotation=ClassificationAnnotation(label_name=label_name)
                    )
                )
            elif isinstance(sample, (tuple, list)) and isinstance(sample[0], Image.Image):
                image_pil, label_name = sample
                typed_samples.append(
                    ImageClassificationSample(
                        data=self._encoder.to_base64(image_pil),
                        annotation=ClassificationAnnotation(label_name=label_name),
                    )
                )
            elif isinstance(sample, ImageClassificationSample):
                typed_samples.append(sample)
            else:
                raise ValueError(f"Unknown sample type: {type(sample)}")
        return typed_samples

    def _create_labels_as_needed(self, samples: List[ImageClassificationSample]) -> None:
        existing_labels = self._label_handler.list_labels(None)
        existing_label_names = {label.name for label in existing_labels}
        new_label_names = {sample.annotation.label_name for sample in samples if sample.annotation}
        missing_label_names = new_label_names - existing_label_names
        missing_labels = [ClassificationLabel(name=label_name) for label_name in missing_label_names]
        if len(missing_labels) > 0:
            self._label_handler.create_labels(missing_labels)

    def _strip_label_names(self, samples: List[ImageClassificationSample]) -> List[ImageClassificationSample]:
        for sample in samples:
            if sample.annotation:
                sample.annotation.label_name = sample.annotation.label_name.strip()
        return samples
